#!/usr/bin/env python3
import aio_pika
import sys
import json
import socket
import asyncio
import pathlib
from importlib import import_module, invalidate_caches
from functools import partial

credentials = None
connection_params = None
hostname = ""
exchange = None


def import_all_c2_functions():
    import glob

    # Get file paths of all modules.
    modules = glob.glob("c2_functions/*.py")
    invalidate_caches()
    for x in modules:
        if not x.endswith("__init__.py") and x[-3:] == ".py":
            module = import_module("c2_functions." + pathlib.Path(x).stem, package=None)
            for el in dir(module):
                if "__" not in el:
                    globals()[el] = getattr(module, el)


async def rabbit_c2_rpc_callback(
    exchange: aio_pika.Exchange, message: aio_pika.IncomingMessage
):
    with message.process():
        try:
            request = json.loads(message.body.decode())
            response = await globals()[request["action"]](request)
            if request["action"] == "translate_from_c2_format":
                response = json.dumps(response).encode()
        except Exception as e:
            print(str(e))
            sys.stdout.flush()
            response = b""
        try:
            await exchange.publish(
                aio_pika.Message(body=response, correlation_id=message.correlation_id),
                routing_key=message.reply_to,
            )
        except Exception as e:
            print(
                "[-] Exception trying to send message back to container for rpc! " + str(e)
            )
            sys.stdout.flush()


async def mythic_service():
    global hostname
    global exchange
    connection = None
    config_file = open("rabbitmq_config.json", "rb")
    main_config = json.loads(config_file.read().decode("utf-8"))
    config_file.close()
    if main_config["name"] == "hostname":
        hostname = socket.gethostname()
    else:
        hostname = main_config["name"]
    import_all_c2_functions()
    while connection is None:
        try:
            connection = await aio_pika.connect_robust(
                host=main_config["host"],
                login=main_config["username"],
                password=main_config["password"],
                virtualhost=main_config["virtual_host"],
            )
            channel = await connection.channel()
            # get a random queue that only the apfell server will use to listen on to catch all heartbeats
            queue = await channel.declare_queue("{}_rpc_queue".format(hostname), auto_delete=True)
            await channel.set_qos(prefetch_count=50)
            try:
                task = queue.consume(
                    partial(rabbit_c2_rpc_callback, channel.default_exchange)
                )
                result = await asyncio.wait_for(task, None)
            except Exception as e:
                print("[-] Exception in connect_and_consume .consume: {}\n trying again...".format(str(e)))
                sys.stdout.flush()
        except (ConnectionError, ConnectionRefusedError) as c:
            print("[-] Connection to rabbitmq failed, trying again...")
            sys.stdout.flush()
        except Exception as e:
            print("[-] Exception in connect_and_consume_rpc connect: {}\n trying again...".format(str(e)))
            # print("Exception in connect_and_consume connect: {}".format(str(e)))
            sys.stdout.flush()
        await asyncio.sleep(2)

def start_service():
    # start our service
    loop = asyncio.get_event_loop()
    loop.create_task(mythic_service())
    loop.run_forever()



from . import MythicBaseRPC
import base64


class MythicRPCResponse(MythicBaseRPC.RPCResponse):
    def __init__(self, resp: MythicBaseRPC.RPCResponse):
        super().__init__(resp._raw_resp)
        if resp.status == MythicBaseRPC.MythicStatus.Success:
            self.data = resp.response
        else:
            self.data = None

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, data):
        self._data = data


class MythicCallbackRPC(MythicBaseRPC.MythicBaseRPC):
    # returns dictionary of `{"raw": raw_tasking, "encrypted": base64(uuid+encrypted_tasking)}`
    async def get_tasking(
        self, uuid: str, tasking_size: int = 1
    ) -> MythicRPCResponse:
        resp = await self.call(
            {
                "action": "get_tasking",
                "uuid": uuid,
                "tasking_size": tasking_size,
            }
        )
        return MythicRPCResponse(resp)

    async def add_route(
        self,
        source_uuid: str,
        destination_uuid: str,
        direction: int = 1,
        metadata: str = None,
    ) -> MythicRPCResponse:
        resp = await self.call(
            {
                "action": "add_route",
                "source": source_uuid,
                "destination": destination_uuid,
                "direction": direction,
                "metadata": metadata,
            }
        )
        return MythicRPCResponse(resp)

    async def remove_route(
        self,
        source_uuid: str,
        destination_uuid: str,
        direction: int = 1,
        metadata: str = None,
    ) -> MythicRPCResponse:
        resp = await self.call(
            {
                "action": "remove_route",
                "source": source_uuid,
                "destination": destination_uuid,
                "direction": direction,
                "metadata": metadata,
            }
        )
        return MythicRPCResponse(resp)

    async def get_callback_info(self, uuid: str) -> MythicRPCResponse:
        resp = await self.call({"action": "get_callback_info", "uuid": uuid})
        return MythicRPCResponse(resp)

    async def get_encryption_data(self, uuid: str, profile: str) -> MythicRPCResponse:
        resp = await self.call(
            {
                "action": "get_encryption_data",
                "uuid": uuid,
                "c2_profile": profile,
            }
        )
        return MythicRPCResponse(resp)

    async def update_callback_info(self, uuid: str, info: dict) -> MythicRPCResponse:
        resp = await self.call(
            {"action": "update_callback_info", "uuid": uuid, "data": info}
        )
        return MythicRPCResponse(resp)

    async def add_event_message(
        self, message: str, level: str = "info"
    ) -> MythicRPCResponse:
        resp = await self.call(
            {"action": "add_event_message", "level": level, "message": message}
        )
        return MythicRPCResponse(resp)

    async def encrypt_bytes(
        self, data: bytes, uuid: str, with_uuid: bool = False,
    ) -> MythicRPCResponse:
        resp = await self.call(
            {
                "action": "encrypt_bytes",
                "data": base64.b64encode(data).decode(),
                "uuid": uuid,
                "with_uuid": with_uuid,
            }
        )
        return MythicRPCResponse(resp)

    async def decrypt_bytes(
        self, data: bytes, uuid: str, with_uuid: bool = False,
    ) -> MythicRPCResponse:
        resp = await self.call(
            {
                "action": "decrypt_bytes",
                "uuid": uuid,
                "data": base64.b64encode(data).decode(),
                "with_uuid": with_uuid,
            }
        )
        return MythicRPCResponse(resp)



import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="mythic_translator_container",
    version="0.0.1",
    description="Functionality for Mythic C2 Translator Containers",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://docs.mythic-c2.net/customizing/c2-related-development",
    author="@its_a_feature_",
    author_email="",
    license="BSD3",
    classifiers=[
        "License :: Other/Proprietary License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=["mythic_translator_container"],
    include_package_data=True,
    install_requires=["aio_pika"],
    entry_points={
    },
)
